#
# makefile.bas
#
# rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/makefile.bas,v 1.1 1994/11/01 03:51:43 tuecke Exp $"
#

O_DIR=.

OBJS = \
        $(O_DIR)/args.$(G_OFILE) \
        $(O_DIR)/error.$(G_OFILE) \
        $(O_DIR)/fileio.$(G_OFILE) \
        $(O_DIR)/init.$(G_OFILE) \
        $(O_DIR)/mac_func.$(G_OFILE) \
        $(O_DIR)/malloc.$(G_OFILE) \
        $(O_DIR)/sanity.$(G_OFILE)  \
        $(O_DIR)/trace.$(G_OFILE)  \
        $(O_DIR)/util.$(G_OFILE)  \
        $(O_DIR)/$(G_PORTS0_TH_MODULE).$(G_OFILE)

HDRS = \
	ports0_config.h \
	$(SRC_PATH)/defs.h \
	$(SRC_PATH)/funcs.h \
	$(SRC_PATH)/globals.h \
	$(SRC_PATH)/internal.h \
	$(SRC_PATH)/macros.h \
	$(SRC_PATH)/ports0.h \
	$(SRC_PATH)/p0_c_hdr.h \
	$(SRC_PATH)/sanity.h \
	$(SRC_PATH)/$(G_PORTS0_TH_MODULE).h \
	$(SRC_PATH)/$(G_PORTS0_TH_MODULE)_i.h \
	$(SRC_PATH)/types.h

INSTALL_HDRS = \
	ports0_config.h \
	$(SRC_PATH)/ports0.h \
	$(SRC_PATH)/p0_c_hdr.h \
	$(SRC_PATH)/sanity.h \
	$(SRC_PATH)/$(G_PORTS0_TH_MODULE).h 

LIBBASE = libports0
LIB = $(LIBBASE)$(FLAVOR_SUFFIX).a

OPT_FLAVOR = opt
SAFE_FLAVOR = safe
DEBUG_FLAVOR = debug
PROFILE_FLAVOR = profile

OPT_SUFFIX = _$(OPT_FLAVOR)
SAFE_SUFFIX = _$(SAFE_FLAVOR)
DEBUG_SUFFIX = _$(DEBUG_FLAVOR)
PROFILE_SUFFIX = _$(PROFILE_FLAVOR)

#FLAVOR_SUFFIXES = $(DEBUG_SUFFIX)
#FLAVOR_SUFFIXES = $(OPT_SUFFIX) $(DEBUG_SUFFIX)
FLAVOR_SUFFIXES = $(OPT_SUFFIX) $(DEBUG_SUFFIX) \
			$(PROFILE_SUFFIX) $(SAFE_SUFFIX)

OPT_DIR = $(OPT_FLAVOR)
SAFE_DIR = $(SAFE_FLAVOR)
DEBUG_DIR = $(DEBUG_FLAVOR)
PROFILE_DIR = $(PROFILE_FLAVOR)

MY_CFLAGS= $(G_CFLAGS) -I$(SRC_PATH) -I$(INCLUDE_PATH) -I. \
		$(FLAVOR_CFLAGS)

default:	build

build:		lib

force:

clean:	force
	- $(G_RM) -rf ports0_config.h *.o $(LIBBASE)* \
		$(OPT_DIR) $(SAFE_DIR) $(DEBUG_DIR) $(PROFILE_DIR) \
		test testo testp tests testo.c testp.c tests.c \
		core a.out *.log *~ \#*

objs:	$(OBJS)

lib: force	
	for i in $(FLAVOR_SUFFIXES) ; \
	  do \
	    $(G_MAKE) build$${i} ; \
	    if [ $$? -ne 0 ] ; then exit 1 ; fi \
	  done

build_opt: force
	- if [ ! -d $(OPT_DIR) ] ; then $(G_MKDIR) $(OPT_DIR) ; fi
	$(G_MAKE) $(LIBBASE)_opt.a \
		"O_DIR=$(OPT_DIR)" \
		"FLAVOR_SUFFIX=_opt" \
		"FLAVOR_CFLAGS=$(G_OPT_CFLAGS)"

build_safe: force
	- if [ ! -d $(SAFE_DIR) ] ; then $(G_MKDIR) $(SAFE_DIR) ; fi
	$(G_MAKE) $(LIBBASE)_safe.a \
		"O_DIR=$(SAFE_DIR)" \
		"FLAVOR_SUFFIX=_safe" \
		"FLAVOR_CFLAGS=$(G_SAFE_CFLAGS)"

build_debug: force
	- if [ ! -d $(DEBUG_DIR) ] ; then $(G_MKDIR) $(DEBUG_DIR) ; fi
	$(G_MAKE) $(LIBBASE)_debug.a \
		"O_DIR=$(DEBUG_DIR)" \
		"FLAVOR_SUFFIX=_debug" \
		"FLAVOR_CFLAGS=$(G_DEBUG_CFLAGS)"

build_profile: force
	- if [ ! -d $(PROFILE_DIR) ] ; then $(G_MKDIR) $(PROFILE_DIR) ; fi
	$(G_MAKE) $(LIBBASE)_profile.a \
		"O_DIR=$(PROFILE_DIR)" \
		"FLAVOR_SUFFIX=_profile" \
		"FLAVOR_CFLAGS=$(G_PROFILE_CFLAGS)"


install: install_libs install_headers

install_libs:
	for i in $(FLAVOR_SUFFIXES) ; \
	  do \
	    $(G_MAKE) "INST_LIB=$(LIBBASE)$${i}.a" install_lib ; \
	    if [ $$? -ne 0 ] ; then exit 1 ; fi \
	  done

install_lib: force
	$(G_CP) $(INST_LIB) $(INSTALL_LIB_DIR)/$(INST_LIB)
	$(G_CHMOD) $(G_FILE_MODE) $(INSTALL_LIB_DIR)/$(INST_LIB)
	$(G_RANLIB) $(INSTALL_LIB_DIR)/$(INST_LIB)

install_headers: force
	for hdr in $(INSTALL_HDRS) ; do \
	  if [ -f $${hdr} ] ; then \
	    cp $${hdr} $(INSTALL_INC_DIR)/. ; \
	  fi \
	done


$(LIB): $(OBJS)
	$(G_RM) -f $(LIB)
	$(G_AR) ruv $(LIB) $(OBJS)
	$(G_RANLIB)  $(LIB)


$(O_DIR)/args.$(G_OFILE):	$(SRC_PATH)/args.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/args.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./args.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/error.$(G_OFILE):	$(SRC_PATH)/error.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/error.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./error.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/fileio.$(G_OFILE):	$(SRC_PATH)/fileio.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/fileio.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./fileio.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/init.$(G_OFILE):	$(SRC_PATH)/init.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/init.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./init.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/mac_func.$(G_OFILE):	$(SRC_PATH)/mac_func.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/mac_func.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./mac_func.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/malloc.$(G_OFILE):	$(SRC_PATH)/malloc.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/malloc.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./malloc.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/sanity.$(G_OFILE):	$(SRC_PATH)/sanity.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/sanity.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./sanity.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/trace.$(G_OFILE):	$(SRC_PATH)/trace.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/trace.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./trace.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/util.$(G_OFILE):	$(SRC_PATH)/util.c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/util.c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./util.$(G_OFILE) $(O_DIR)/. ; \
	fi

$(O_DIR)/$(G_PORTS0_TH_MODULE).$(G_OFILE):	$(SRC_PATH)/$(G_PORTS0_TH_MODULE).c $(HDRS)
	$(G_CC) $(MY_CFLAGS) -c $(SRC_PATH)/$(G_PORTS0_TH_MODULE).c
	- if [ $(O_DIR) != "." ] ; then \
		$(G_MV) ./$(G_PORTS0_TH_MODULE).$(G_OFILE) $(O_DIR)/. ; \
	fi


ports0_config.h:	makefile makefile.bas
	-$(G_RM) -f ports0_config.h
	@echo "Creating ports0_config.h"
	echo "" > ports0_config.h ; \
	echo "#ifndef _PORTS0_INCLUDE_PORTS0_CONFIG_H" >> ports0_config.h ; \
	echo "#define _PORTS0_INCLUDE_PORTS0_CONFIG_H" >> ports0_config.h ; \
	echo "" >> ports0_config.h ; \
	echo "#define PORTS0_ARCH_$(G_ARCH)" >> ports0_config.h ; \
	for i in $(G_OTHER_ARCH) ; \
	  do \
	    echo "#define PORTS0_ARCH_$$i" >> ports0_config.h ; \
	done; \
	echo "#define PORTS0_ARCH_STRING \"$(G_ARCH)\"" >> ports0_config.h ; \
	echo "#define PORTS0_ARCH_TYPE PORTS0_ARCH_TYPE_$(G_ARCH)" >> ports0_config.h ; \
	echo "" >> ports0_config.h ; \
	echo "#define PORTS0_TH_HEADER \"$(G_PORTS0_TH_MODULE).h\"" >> ports0_config.h ; \
	echo "#define PORTS0_THI_HEADER \"$(G_PORTS0_TH_MODULE)_i.h\"" >> ports0_config.h ; \
	echo "" >> ports0_config.h ; \
	for i in $(G_PORTS0_CONFIG) ; \
	  do \
	    out1=`echo $$i | sed s/=.*$$//` ; \
	    out2=`echo $$i | sed s/$${out1}=\*//` ; \
	    echo "#ifndef $$out1" >> ports0_config.h ; \
	    echo "#define $$out1 $$out2" >> ports0_config.h ; \
	    echo "#endif" >> ports0_config.h ; \
	    echo "" >> ports0_config.h ; \
	done; \
	echo "#endif /* _PORTS0_INCLUDE_PORTS0_CONFIG_H */" >> ports0_config.h ; \
	echo "" >> ports0_config.h 


ports0.tar: force
	tar cf ports0.tar *.c *.h makefile.bas

ports0.tar.Z: force
	tar cf - *.c *.h makefile.bas | compress > ports0.tar.Z


test:	build_$(DEBUG_FLAVOR) test.$(G_OFILE)
	$(G_RM) -f test
	$(G_CC) $(G_DEBUG_CFLAGS) $(G_LDFLAGS) -o test test.$(G_OFILE) \
		$(LIBBASE)_$(DEBUG_FLAVOR).a $(G_CLIBS)

test.$(G_OFILE):	$(SRC_PATH)/test.c $(SRC_PATH)/ports0.h
	$(G_CC) $(MY_CFLAGS) $(G_DEBUG_CFLAGS) \
		-c $(SRC_PATH)/test.c


testo:	build_$(OPT_FLAVOR) testo.$(G_OFILE)
	$(G_RM) -f testo
	$(G_CC) $(G_OPT_CFLAGS) $(G_LDFLAGS) -o testo testo.$(G_OFILE) \
		$(LIBBASE)_$(OPT_FLAVOR).a $(G_CLIBS)

testo.$(G_OFILE):	$(SRC_PATH)/testo.c $(SRC_PATH)/ports0.h
	$(G_CC) $(MY_CFLAGS) $(G_OPT_CFLAGS) \
		-DTEST_EXECUTABLE_NAME=\"testo\" \
		-c $(SRC_PATH)/testo.c


testp:	build_$(PROFILE_FLAVOR) testp.$(G_OFILE)
	$(G_RM) -f testp
	$(G_CC) $(G_PROFILE_CFLAGS) $(G_LDFLAGS) -o testp testp.$(G_OFILE) \
		$(LIBBASE)_$(PROFILE_FLAVOR).a $(G_CLIBS)

testp.$(G_OFILE):	$(SRC_PATH)/testp.c $(SRC_PATH)/ports0.h
	$(G_CC) $(MY_CFLAGS) $(G_PROFILE_CFLAGS) \
		-DTEST_EXECUTABLE_NAME=\"testp\" \
		-c $(SRC_PATH)/testp.c


tests:	build_$(SAFE_FLAVOR) tests.$(G_OFILE)
	$(G_RM) -f tests
	$(G_CC) $(G_SAFE_CFLAGS) $(G_LDFLAGS) -o tests tests.$(G_OFILE) \
		$(LIBBASE)_$(SAFE_FLAVOR).a $(G_CLIBS)

tests.$(G_OFILE):	$(SRC_PATH)/tests.c $(SRC_PATH)/ports0.h
	$(G_CC) $(MY_CFLAGS) $(G_SAFE_CFLAGS) \
		-DTEST_EXECUTABLE_NAME=\"tests\" \
		-c $(SRC_PATH)/tests.c


tc:
	$(G_RM) -f testo.c testp.c tests.c
	$(G_SYMLINK) test.c testo.c
	$(G_SYMLINK) test.c testp.c
	$(G_SYMLINK) test.c tests.c
