/*
 * ports0.h
 *
 * rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/ports0.h,v 1.1 1994/11/01 03:52:22 tuecke Exp $"
 *
 * This header contains the exported interface of Ports0.
 */

#ifndef _PORTS0_INCLUDE_PORTS0_H
#define _PORTS0_INCLUDE_PORTS0_H

/*
 * EXTERN_C_BEGIN and EXTERN_C_END should surround any Ports0 prototypes in
 * ports0.h or Ports0 .h files that are included by ports0.h.  This will
 * allow C++ codes to include ports0.h and properly link against the
 * Ports0 library.
 */
#ifdef __cplusplus
#define EXTERN_C_BEGIN extern "C" {
#define EXTERN_C_END }
#else
#define EXTERN_C_BEGIN
#define EXTERN_C_END
#endif

EXTERN_C_BEGIN

/*
 * ports0_arch_type_t
 */
typedef enum _ports0_arch_type_t
{
    PORTS0_ARCH_TYPE_UNKNOWN,
    PORTS0_ARCH_TYPE_ANY,
    PORTS0_ARCH_TYPE_SUN3,
    PORTS0_ARCH_TYPE_SUN4,
    PORTS0_ARCH_TYPE_NEXT040,
    PORTS0_ARCH_TYPE_NEXTX86,
    PORTS0_ARCH_TYPE_NEXTPA,
    PORTS0_ARCH_TYPE_NEXTSPARC,
    PORTS0_ARCH_TYPE_RS6000,
    PORTS0_ARCH_TYPE_PARAGON,
    PORTS0_ARCH_TYPE_NTX86,
    PORTS0_ARCH_TYPE_CRAYC90
} ports0_arch_type_t;

EXTERN_C_END

#include "ports0_config.h"

#ifndef PORTS0_ARCH_TYPE
#define PORTS0_ARCH_TYPE PORTS0_ARCH_TYPE_UNKNOWN
#endif

#include "p0_c_hdr.h"

EXTERN_C_BEGIN

#if !defined(symmetry) || !defined(_TYPES_)
/* Only do this typedef if sys/types.h has not been included */
typedef int		bool_t;
#endif

#define	PORTS0_TRUE		1
#define	PORTS0_FALSE		0

#ifndef NULL
#define NULL ((void *) 0)
#endif

#define PORTS0_MAX(V1,V2) (((V1) > (V2)) ? (V1) : (V2))
#define PORTS0_MIN(V1,V2) (((V1) < (V2)) ? (V1) : (V2))

#ifdef PORTS0_DEFINE_GLOBALS
#define PORTS0_GLOBAL
#else
#define PORTS0_GLOBAL extern
#endif


#ifdef PORTS0_DEBUG
#define Ports0Assert(assertion) \
    if (!(assertion)) \
    { \
        ports0_fatal("Assertion " #assertion " failed in file %s at line %d\n",\
		   __FILE__, __LINE__); \
    }
#define Ports0Assert2(assertion, print_args) \
    if (!(assertion)) \
    { \
        ports0_fatal("Assertion " #assertion " failed in file %s at line %d: %s", \
		   __FILE__, __LINE__, ports0_assert_sprintf print_args); \
    }
#else /* PORTS0_DEBUG */
#define Ports0Assert(assertion)
#define Ports0Assert2(assertion, print_args)
#endif /* PORTS0_DEBUG */


/*
 * Some typedefs...
 */
typedef void *(*ports0_thread_func_t)(void *user_arg);

EXTERN_C_END

/*
 * Sanity checking inspects the data structures and
 * tries to determine if corruption has occurred
 * before the system would crash.  Threads then
 * can be "imprisoned" so that they may be inspected.
 * Important since multithreaded applications can
 * be non-deterministic.
 * sanity.h is placed into ports0.h because the macros
 * that are here should be available to the user,
 * since they are called from macros that the user
 * calls.
 * Sanity checking is turned on/off by defining
 * PORTS0_SANITY_CHECK
 * JMP 02/28/94
 */

#include "sanity.h"

/*
 * Include the thread module's external header, which should contain:
 *   1) #include's for system thread headers
 *   2) typedefs for external ports0 types
 *   3) macros for external ports0 routines
 */
#ifndef PORTS0_TH_HEADER
#define PORTS0_TH_HEADER "th.h"
#endif

#include PORTS0_TH_HEADER


EXTERN_C_BEGIN

/*********************************************************************
 * 		Exported interface declarations
 *********************************************************************/
extern void     ports0_init (int *argc,
			     char **argv[],
			     char *package_id);

extern void	ports0_shutdown();


/*
 * Reentrant lock
 */
PORTS0_GLOBAL ports0_mutex_t ports0_reentrant_mutex;
#define ports0_macro_reentrant_lock() \
    ports0_mutex_lock(&ports0_reentrant_mutex)
#define ports0_macro_reentrant_unlock() \
    ports0_mutex_unlock(&ports0_reentrant_mutex)
#ifndef PORTS0_USE_MACROS
#define ports0_reentrant_lock()   ports0_macro_reentrant_lock()
#define ports0_reentrant_unlock() ports0_macro_reentrant_unlock()
#else  /* PORTS0_USE_MACROS */
extern int ports0_reentrant_lock(void);
extern int ports0_reentrant_unlock(void);
#endif /* PORTS0_USE_MACROS */

#ifdef PORTS0_STDIO_NOT_REENTRANT
#define ports0_stdio_lock() ports_reentrant_lock()
#define ports0_stdio_unlock() ports_reentrant_unock()
#else
#define ports0_stdio_lock()
#define ports0_stdio_unlock()
#endif

/*
 * Malloc #defines.
 * Must handle three cases:
 *    1) PORTS0_MALLOC_DEBUG: Call ports0_debug_*() routines.
 *    2) PORTS0_MALLOC_NOT_REENTRANT: Call ports0_reentrant_*() routines.
 *    3) Else: Call underlying memory allocation routines.
 */
#ifdef PORTS0_MALLOC_DEBUG

/* This doesn't work yet */
extern void *		ports0_debug_malloc(int bytes, char *file, int line);
extern void		ports0_debug_malloc_check(char *file, int line);
extern void		ports0_debug_show_freed_blocks();
extern void		ports0_debug_show_malloc_stats();
extern void		ports0_debug_free(void *ptr, char *file, int line);
extern bool_t		ports0_malloc_debug_level();

#define ports0_macro_malloc(bytes) \
    ports0_debug_malloc(bytes,__FILE__,__LINE__)
#define ports0_macro_realloc(ptr,bytes) \
    ports0_debug_realloc(ptr,bytes,__FILE__,__LINE__)
#define ports0_macro_calloc(nobj,bytes) \
    ports0_debug_calloc(nobj,bytes,__FILE__,__LINE__)
#define ports0_macro_free(ptr) \
    ports0_debug_free(ptr,__FILE__,__LINE__)

#else /* PORTS0_MALLOC_DEBUG */

#ifdef PORTS0_MALLOC_NOT_REENTRANT

#define ports0_macro_malloc(bytes) \
    ports0_reentrant_malloc(bytes)
#define ports0_macro_realloc(ptr,bytes) \
    ports0_reentrant_realloc(ptr,bytes)
#define ports0_macro_calloc(nobj,bytes) \
    ports0_reentrant_calloc(nobj,bytes)
#define ports0_macro_free(ptr) \
    ports0_reentrant_free(ptr)

#else  /* PORTS0_MALLOC_NOT_REENTRANT */

#define ports0_macro_malloc(bytes) \
    malloc(bytes)
#define ports0_macro_realloc(ptr,bytes) \
    realloc(ptr,bytes)
#define ports0_macro_calloc(nobj,bytes) \
    calloc(nobj,bytes)
#define ports0_macro_free(ptr) \
    free(ptr)

#endif /* PORTS0_MALLOC_NOT_REENTRANT */

#endif /* PORTS0_MALLOC_DEBUG */

/*
 * Memory allocation routines
 */
#ifndef NEXUS_USE_MACROS
extern void *ports0_malloc(size_t bytes);
extern void *ports0_realloc(void *ptr,
			    size_t bytes);
extern void *ports0_calloc(size_t nobj, 
			   size_t bytes);
extern void ports0_free(void *ptr);
#else /* NEXUS_USE_MACROS */
#define ports0_malloc(bytes) \
    ports0_macro_malloc(bytes)
#define ports0_realloc(ptr,bytes) \
    ports0_macro_realloc(ptr,bytes)
#define ports0_calloc(nobj,bytes) \
    ports0_macro_calloc(nobj,bytes)
#define ports0_free(data_segment) \
    ports0_macro_free(data_segment)
#endif /* NEXUS_USE_MACROS */


#define Ports0Malloc(Func, Var, Type, Size) \
{ \
    if ((Size) > 0) \
    { \
	if (((Var) = (Type) ports0_macro_malloc (Size)) == (Type) NULL) \
	{ \
	    ports0_fatal("%s: malloc of size %d failed for %s %s in file %s line %d\n", \
                        #Func, (Size), #Type, #Var, __FILE__, __LINE__); \
	} \
    } \
    else \
    { \
	(Var) = (Type) NULL; \
    } \
}

#define Ports0Free(Ptr) \
{ \
    if ((Ptr) != NULL) \
    { \
	ports0_macro_free(Ptr); \
    } \
}

/*
 * File I/O routines
 */
#ifdef PORTS0_FILEIO_NO_REENTRANT

#define ports0_macro_open(path, flags, mode) \
    ports0_reentrant_open(path, flags, mode)
#define ports0_macro_close(fd) \
    ports0_reentrant_close(fd)
#define ports0_macro_read(fd, buf, nbytes) \
    ports0_reentrant_read(fd, buf, nbytes)
#define ports0_macro_write(fd, buf, nbytes) \
    ports0_reentrant_write(fd, buf, nbytes)
#define ports0_macro_fstat(fd, buf) \
    ports0_reentrant_fstat(fd, buf)
#define ports0_macro_lseek(fd, offset, whence) \
    ports0_reentrant_lseek(fd, offset, whence)

#else  /* PORTS0_FILEIO_NO_REENTRANT */

#define ports0_macro_open(path, flags, mode) \
        open(path, flags, mode)
#define ports0_macro_close(fd) \
        close(fd)
#define ports0_macro_read(fd, buf, nbytes) \
        read(fd, buf, nbytes)
#define ports0_macro_write(fd, buf, nbytes) \
        write(fd, buf, nbytes)
#define ports0_macro_fstat(fd, buf) \
        fstat(fd, buf)
#define ports0_macro_lseek(fd, offset, whence) \
        lseek(fd, offset, whence)

#endif /* PORTS0_FILEIO_NO_REENTRANT */

#ifndef PORTS0_USE_MACROS
extern int ports0_open(char *path,
		       int flags,
		       int mode);
extern int ports0_close(int fd);
extern int ports0_read(int fd,
		       char *buf,
		       int nbytes);
extern int ports0_write(int fd,
			char *buf,
			int nbytes);
extern int ports0_fstat(int fd,
			struct stat *buf);
extern int ports0_lseek(int fd,
			off_t offset,
			int whence);
#else  /* PORTS0_USE_MACROS */
#define ports0_open(path, flags, mode) \
    ports0_macro_open(path, flags, mode)
#define ports0_close(fd) \
    ports0_macro_close(fd)
#define ports0_read(fd, buf, nbytes) \
    ports0_macro_read(fd, buf, nbytes)
#define ports0_write(fd, buf, nbytes) \
    ports0_macro_write(fd, buf, nbytes)
#define ports0_fstat(fd, buf) \
    ports0_macro_fstat(fd, buf)
#define ports0_lseek(fd, offset, whence) \
    ports0_macro_lseek(fd, offset, whence)
#endif /* PORTS0_USE_MACROS */



/*********************************************************************
 * Other exported interface declarations that are not
 * officially part of the ports0 interface.
 *********************************************************************/
/*
 * args.c
 */
extern void		ports0_usage_message();
extern int		ports0_new_process_params(char *buf, int size);

/*
 * error.c
 */
extern void		ports0_silent_fatal();
#ifdef __STDC__
extern void		ports0_fatal(char *msg, ...);
extern void		ports0_error(char *msg, ...);
extern void		ports0_warning(char *msg, ...);
extern void		ports0_notice(char *msg, ...);
extern void		ports0_printf(char *msg, ...);
extern void		ports0_perror(char *msg, ...);
extern void		ports0_fatal_perror(char *msg, ...);
extern char *		ports0_assert_sprintf(char *msg, ...);
#else /* __STDC__ */
extern void		ports0_fatal();
extern void		ports0_error();
extern void		ports0_warning();
extern void		ports0_notice();
extern void		ports0_printf();
extern void		ports0_perror();
extern void		ports0_fatal_perror();
extern char *		ports0_assert_sprintf();
#endif /* __STDC__ */


/*
 * th_*.c
 */
extern bool_t		ports0_preemptive_threads();

EXTERN_C_END

#endif /* _PORTS0_INCLUDE_PORTS0_H */
