/*
 * th_c.h
 *
 * External header for a C-threads based Ports0 threads library.
 * This will be included by user code.
 *
 * rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_c.h,v 1.1 1994/11/01 03:53:01 tuecke Exp $"
 */

#include <mach/cthreads.h>

typedef struct mutex		ports0_mutex_t;
typedef struct condition	ports0_cond_t;

typedef int			ports0_thread_attr_t;
typedef int			ports0_mutexattr_t;
typedef int			ports0_condattr_t;

typedef void *			ports0_thread_t;

#define PORTS0_THREAD_ONCE_INIT 0
typedef int ports0_thread_once_t;

typedef int ports0_thread_key_t;
typedef void (*ports0_thread_key_destructor_func_t)(void *value);


/*
 * Some externs...
 */
extern int		ports0_thread_create(ports0_thread_t *thread,
					    ports0_thread_attr_t *attr,
					    ports0_thread_func_t func,
					    void *user_arg);
extern void		ports0_thread_exit(void *status);

extern int		ports0_thread_key_create(ports0_thread_key_t *key,
				ports0_thread_key_destructor_func_t func);
extern void		ports0_thread_setspecific(ports0_thread_key_t key,
						 void *value);
extern void		ports0_thread_getspecific(ports0_thread_key_t key,
						 void **value);


/*
 * macros and prototypes for thread manipulation routines
 */
#define ports0_macro_thread_yield() \
    cthread_yield()
#define ports0_macro_thread_self() \
    ((void *) (cthread_self()))
#define ports0_macro_thread_equal(t1,t2) \
    (t1 == t2)
#define ports0_macro_thread_once(once_control, init_routine) \
    (*once_control ? 0 : _p0_actual_thread_once(once_control, init_routine))
#ifndef PORTS0_USE_MACROS
extern void		ports0_thread_yield();
extern ports0_thread_t	ports0_thread_self(void);
extern int		ports0_thread_equal(ports0_thread_t t1,
					   ports0_thread_t t2);
extern int		ports0_thread_once(ports0_thread_once_t *once_control,
					  void (*init_routine)());
#else  /* PORTS0_USE_MACROS */
#define ports0_thread_yield() \
    ports0_macro_thread_yield()
#define ports0_thread_self() \
    ports0_macro_thread_self()
#define ports0_thread_equal(t1,t2) \
    ports0_macro_thread_equal(t1,t2)
#define ports0_thread_once(once_control, init_routine) \
    ports0_macro_thread_once(once_control, init_routine)
#endif /* PORTS0_USE_MACROS */


/*
 * lock macros and prototypes
 */
#define ports0_macro_mutex_init(mutex, attr)	mutex_init(mutex)
#define ports0_macro_mutex_destroy(mutex)	mutex_clear(mutex)
#define ports0_macro_mutex_lock(mutex)		mutex_lock(mutex)
#define ports0_macro_mutex_unlock(mutex)	mutex_unlock(mutex)

#ifndef PORTS0_USE_MACROS
extern void		ports0_mutex_init(ports0_mutex_t *mutex,
					  ports0_mutexattr_t *attr);
extern void		ports0_mutex_destroy(ports0_mutex_t *mutex);
extern void		ports0_mutex_lock(ports0_mutex_t *mutex);
extern void		ports0_mutex_unlock(ports0_mutex_t *mutex);
#else  /* PORTS0_USE_MACROS */
#define ports0_mutex_init(mutex, attr) \
    ports0_macro_mutex_init(mutex, attr)
#define ports0_mutex_destroy(mutex) \
    ports0_macro_mutex_destroy(mutex)
#define ports0_mutex_lock(mutex) \
    ports0_macro_mutex_lock(mutex)
#define ports0_mutex_unlock(mutex) \
    ports0_macro_mutex_unlock(mutex)
#endif /* PORTS0_USE_MACROS */


/*
 * condition macros and prototypes
 */
#define ports0_macro_cond_init(cond, attr) \
    condition_init(cond)
#define ports0_macro_cond_destroy(cond) \
    condition_clear(cond)
#define ports0_macro_cond_wait(cond, mutex) \
    condition_wait(cond, mutex)
#define ports0_macro_cond_signal(cond) \
    condition_signal(cond)
#define ports0_macro_cond_broadcast(cond) \
    condition_broadcast(cond)

#ifndef PORTS0_USE_MACROS
extern void	ports0_cond_init(ports0_cond_t *cond, ports0_condattr_t *attr);
extern void	ports0_cond_destroy(ports0_cond_t *cond);
extern void	ports0_cond_wait(ports0_cond_t *cond,
				ports0_mutex_t *mutex);
extern void	ports0_cond_signal(ports0_cond_t *cond);
extern void	ports0_cond_broadcast(ports0_cond_t *cond);
#else  /* PORTS0_USE_MACROS */
#define ports0_cond_init(cond, attr) \
    ports0_macro_cond_init(cond, attr)
#define ports0_cond_destroy(cond) \
    ports0_macro_cond_destroy(cond)
#define ports0_cond_wait(cond, mutex) \
    ports0_macro_cond_wait(cond, mutex)
#define ports0_cond_signal(cond) \
    ports0_macro_cond_signal(cond)
#define ports0_cond_broadcast(cond) \
    ports0_macro_cond_broadcast(cond)
#endif /* PORTS0_USE_MACROS */
