/*
 * Ports0 Threads built on top of QuickThreads
 *
 * author: Craig Lee at Aerospace.
 * 1994 May
 * Preemption, July 1994
 */

static char *rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_qt.c,v 1.1 1994/11/01 03:53:37 tuecke Exp $";

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <signal.h>
#include "internal.h"
/*
#include "th_qt.h"
*/

/*
 *  Const shared data
 */

static int numProcessors = 1;
static int numQueues;
static int processGroup;
/*static int affinity = FALSE;*/

static ports0_thread_attr_t default_attrs;
static ports0_mutexattr_t default_mutex_attrs; /* initializers ?? */
static ports0_condattr_t default_cond_attrs;

/*
 *  Dynamic shared data
 */

static thread_kernel_t *TKp;

static bool_t		using_handler_thread = PORTS0_FALSE;
static ports0_thread_t   handler_thread;
static bool_t		handler_thread_done = PORTS0_FALSE;
static void *		handler_thread_func(void *arg);
static ports0_mutex_t    handler_mutex;
static ports0_cond_t     handler_cond;

static bool_t		preemptive_threads = PORTS0_FALSE;
static volatile int     timer_hit = 0;
static volatile int     timer_hits = 0;
static volatile int     check_slice_hits = 0;
static int              dont_preempt = 0;
static int              preemption_cnt = 0;
static struct itimerval time_slice_val;

#ifndef PORTS0_THREAD_SAFE_STDIO
ports0_mutex_t ports0_stdio_mutex;
#endif

/*
 *  Processor private data
 */

static PrivateProcessor pP;



/*
 *  Thread specific data key maintainance stuff
 */
static int used_keys;
static ports0_thread_key_destructor_func_t key_destructor[PORTS0_DATAKEYS_MAX];
#ifdef MULTIPROC
static spin_lock_t key_lock;
#endif


void ProcessInit( int id );
void ReadyqInit();
int sched_next_activethread();

ports0_thread_t alloc_thread(void);
void alloc_stack( ports0_thread_t tp, int stacksize );
void only( void *argp, ports0_thread_t t, void *(*funcp)() );
void cleanup();
void *csw_helper();

void time_slicer();
void *time_slice_helper();

/*
 * _p0_thread_process_arguments_init()
 *
 * Initialize any variables that will be used to hold
 * my configuration information extracted from arguments.
 */
void _p0_thread_process_arguments_init()
{
}


/*
 * _p0_thread_process_arguments()
 *
 * If ports0_package_getarg(current_arg) is destined for my module,
 * then extract it into my own variables.
 *
 * Return:	A new current argument.
 */
int _p0_thread_process_arguments(int current_arg, int arg_count)
{
  /*  Could extract #procs and default stack size  */
  return (current_arg);
}


/*
 * _p0_thread_usage_message()
 *
 * Print a usage message for this module's arguments to stdout.
 */
void _p0_thread_usage_message()
{
  printf("Signal preemptions: %d  Check yields: %d\n",
	 preemption_cnt, check_slice_hits );
}


/*
 * _p0_thread_new_process_params()
 *
 * Write any new process parameters for this module into 'buf',
 * up to a maximum of 'size' characters.
 *
 * Return:	The number of characters written into 'buf'.
 */
int _p0_thread_new_process_params(char *buf, int size)
{
  return (0);
}

/*
 * _p0_thread_preinit()
 *
 * If you need to call a thread package initialization routine, then
 * do it here.  This is called as the very first thing in ports0_init().
 */
void _p0_thread_preinit( void )
{
} /* _p0_thread_preinit() */


void _p0_thread_init( void )
{
  int i;

  /*  Do basic initialization  */

  numQueues = numProcessors;
#ifdef PORTS0_ARCH_OSF1
  processGroup = getpgrp();
#else
  processGroup = getpgrp(0);
#endif
  default_attrs.stack_size = PORTS0_STACK_MIN; /* Should be set in _p0_thread_process_arguments */

  TKp = (thread_kernel_t *)nq_malloc( sizeof(thread_kernel_t) );
  TKp->IdleProcs = numProcessors;
  TKp->ActiveThreads = TKp->nThreads = 1;
  ReadyqInit();
  TKp->free_count = 0;
  ThreadQInit( &(TKp->freeQ) );
  for  ( i=0; i<MAXPROCS; i++ )  TKp->idle_vec[i] = 0;

  ProcessInit( 0 );

  /* Initialize thread specific keys */

  used_keys = 0;
  for  ( i=0; i<PORTS0_DATAKEYS_MAX; i++ )  key_destructor[i] = NULL;

  /* Initialize monitor for handler thread */

  ports0_mutex_init( &handler_mutex, NULL );
  ports0_cond_init( &handler_cond, NULL );

#ifndef PORTS0_THREAD_SAFE_STDIO
  ports0_mutex_init( &ports0_stdio_mutex, NULL );
#endif
}

/*
 *  Init process private info
 */
void ProcessInit( int id )
{
  pP.myId = id;
  pP.lastLook = pP.putQ = &(TKp->readyList[pP.myId % numQueues]);

  /*  The main ports0 thread becomes the first activeThread.  It already
      has a stack so we don't need to alloc one.  sp will be set the
      first time this thread switches out.
  */

  pP.activeThread = alloc_thread();
  pP.activeThread->tid = 0;
  pP.activeThread->stack_size = 0;
  pP.activeThread->stack_base = 0;
  pP.activeThread->sp = NULL;

  /*  idleThread is really a misnomer.  idleThread actually represents
      the global handler thread _if_ it gets created.  When created, the global
      handler thread runs as a normal thread with no special handling except
      when executing _p0_thread_shutdown_handler_thread(), i.e., when it's trying
      to shut itself down.  In this case, it just sets the done flag and doesn't block.
  */

  pP.idleThread = NULL;

  /* Setup vtimer */

  if  ( preemptive_threads )  {
    signal( SIGVTALRM, time_slicer );
    time_slice_val.it_interval.tv_sec  = time_slice_val.it_value.tv_sec  = 0;
    time_slice_val.it_interval.tv_usec = time_slice_val.it_value.tv_usec = TIME_SLICE;
    if  ( setitimer( ITIMER_VIRTUAL, &time_slice_val, NULL ) )
      _p0_report_bad_rc( 1, "setitimer failed\n" );
  }

  /*  tmp_affinity done here, if possible and desired */
}


void *csw_helper( qt_t *old_sp, void *old_thr, void *qlock )
{
  ((ports0_thread_t)old_thr)->sp = old_sp;

  /* Do queue unlocking here in multiprocessor version */
  spin_unlock( qlock );

  /* Since we just context-switched, throw away any timer hits
     and allow preemption.
  */
  START_PREEMPTION;
  return NULL;
}


void cleanup( ports0_thread_t self )
{
  int i;

  DONT_PREEMPT;
  /* Don't need to CHECK_SLICE later since we're headed for a csw anyway. */

  for  ( i=0; i<used_keys; i++ )  {
    if  ( key_destructor[i] && self->key_data[i] )
      (*(key_destructor[i]))( self->key_data[i] );
    self->key_data[i] = NULL;
  }

  atomic_dec(TKp->ActiveThreads);
  sched_next_activethread(BLOCKING_SCHED);
  /*  Can't let anybody see me on the freeq until I switch.  */
  FreeQPut_switch( self, pP.activeThread );
  /*  Control never gets here.  */
}


int sched_next_activethread( int must_sched_something )
{
  int n = numQueues;
  ThreadQ *q = pP.lastLook;

  for  (;;)  {

    /*  Scan all ready queues starting with mine.  */

    while  ( 0 < n-- )  {
      spin_lock( &(q->lock) );
      if  ( q->qhead )  {	/* Got a ready thread */
	ThreadGet( pP.activeThread, q );
	spin_unlock(&(q->lock));
	return 1;
      }
      spin_unlock(&(q->lock));
      q = NextQ(q);
    }

    /*  If I can't find any ready threads (and I don't have a global handler
	thread) _and_ I have to schedule something, then poll all the protocol
	modules.  This may or may not unblock a thread so I have to loop and
	start the whole process over.
    */

    if  ( !must_sched_something )  return 0;

    ports0_poll();

  }
}


void ReadyqInit()
{
  int i;

  for  ( i=0; i<numQueues; i++ )  {
    ThreadQInit( &(TKp->readyList[i]) );
  }
  TKp->firstQ = TKp->readyList;
  TKp->lastQ = TKp->readyList + (numQueues - 1);
}


/*
 * Allocate a red-zoned stack segment for a thread.
 * Threads and their stacks are never deallocated;
 * they are retained on a free list for future use.
 */

void alloc_stack( ports0_thread_t tp, int stacksize )
{
  char *dead_page;

  tp->stack_size = stacksize;
  tp->stack_base = (char *)nq_memalign( getpagesize(), stacksize );

  if  ( (int) tp->stack_base == -1 )
    { errno = ENOMEM; _p0_report_bad_rc( 1, "<Stack allocation>" ); }

  tp->sp = QT_SP( tp->stack_base, stacksize );

#ifdef QT_GROW_DOWN
  dead_page = tp->stack_base;
#else
  dead_page = tp->stack_base + stacksize - getpagesize();
#endif

  if  ( mprotect( dead_page, getpagesize(), PROT_NONE ) )
    _p0_report_bad_rc( 1, "Redzoning failed" );
}


ports0_thread_t alloc_thread()
{
  ports0_thread_t newt;
  void **kdp, **kdp_end;

  newt = (ports0_thread_t)nq_malloc(sizeof(ports0_thread_s));
  if  ( newt == NULL )
    { errno = ENOMEM; _p0_report_bad_rc( 1, "<Thread allocation>" ); }

  kdp = newt->key_data;
  kdp_end = kdp + PORTS0_DATAKEYS_MAX;
  while  ( kdp < kdp_end )  *kdp++ = NULL;
  newt->next = NULL;

  return newt;
}


/*  time_slicer runs on the interrupted thread's stack so I can block
    from here.  Since time_slicer (the sig handler) doesn't actually
    return until the thread is rescheduled, time_slice_helper has to
    explicitly reenable SIGVTALRM and restart the timer.
*/

void time_slicer()
{
  ports0_thread_t self = pP.activeThread;

  timer_hits++;
  if  ( dont_preempt )  timer_hit = 1;
  else if  ( sched_next_activethread(NON_BLOCKING_SCHED) )  {
    preemption_cnt++;
    ThreadPut( self, (pP.putQ) );
    QT_BLOCK( time_slice_helper, self, NULL, pP.activeThread->sp );
  }  else  {
    pP.activeThread = self;
  }
}


void *time_slice_helper( qt_t *old_sp, void *old_thr, void *qlock )
{
  ((ports0_thread_t)old_thr)->sp = old_sp;
  sigsetmask( sigblock(0) & ~(sigmask(SIGVTALRM)) );
  if  ( setitimer( ITIMER_VIRTUAL, &time_slice_val, NULL ) )
    _p0_report_bad_rc( 1, "setitimer failed\n" );

  return NULL;
}

/*
 * ports0_thread_create
 */
/*
  Note: the new thread has to inherit the calling threads context.

  FreeQGet is a simple macro that only checks the head of the free queue
for a stack that is at least as big as stk_size.  Hence, the user will
always get a stack as big as requested but if the requested stack size
varies widely and often, this will tend to create more thread headers
and stacks than are absolutely necessary.
*/

int ports0_thread_create( ports0_thread_t *tp,
			 ports0_thread_attr_t *attrp,
			 ports0_thread_func_t funcp,
			 void *argp )
{
  ports0_thread_t newt;
  int stk_size = ((attrp) ? attrp->stack_size : default_attrs.stack_size );

  DONT_PREEMPT;
  FreeQGet( newt, stk_size );
  if  ( newt == NULL )  {
    newt = alloc_thread();
    alloc_stack( newt, stk_size );
  }  else  {
    newt->sp = QT_SP( newt->stack_base, newt->stack_size );
    newt->next = NULL;
  }

  /* Put single arg on stack and move to ready queue */

  newt->sp = QT_ARGS( newt->sp, argp, newt, funcp, only );

  /* Grab my context, and give it to my child */
  ports0_thread_getspecific(_p0_context_key, 
			   &(newt->key_data[_p0_context_key])); 
  ReadyQPut( newt );
  atomic_inc(TKp->ActiveThreads);
  if  ( tp )  *tp = newt;

  CHECK_SLICE;
  return 0;
}


void only( void *argp, ports0_thread_t t, void *(*funcp)() )
{
  t->tid = atomic_inc(TKp->nThreads) - 1;

  /* Call the user function */
  (*funcp)(argp);

  /* It's an error for only to return; cleanup never does. */
  cleanup( t );
}


ports0_thread_t ports0_thread_self( void )
{ return pP.activeThread; }


#ifndef PORTS0_USE_MACROS
int ports0_thread_equal(ports0_thread_t t1, ports0_thread_t t2)
{
	return  (t1 == t2);
}
#endif

int ports0_thread_attr_init( ports0_thread_attr_t *attrp )
{
  if  ( attrp == NULL )  return EINVAL;
  attrp->stack_addr  = default_attrs.stack_addr;
  attrp->stack_size  = default_attrs.stack_size;
  return 0;
}


int ports0_thread_attr_setstacksize( ports0_thread_attr_t *attrp, size_t len )
{
  if  ( (attrp == NULL) || (len < PORTS0_STACK_MIN) )  return EINVAL;
  attrp->stack_size = (int)len;
  return 0;
}


int ports0_thread_attr_getstacksize( ports0_thread_attr_t *attrp, size_t *lenp )
{
  if  ( (attrp == NULL) || (lenp == NULL) )  return EINVAL;
  *lenp = (size_t)(attrp->stack_size);
  return 0;
}


/*
 * _p0_thread_create_handler_thread()
 *
 **  This bunch of routines have been lifted from th_c.c.
 *
 * If we have preemptive threads:
 *   Create a normal priority thread that will sit in a
 *   loop calling ports0_poll().
 *
 * If we have non-preemptive threads:
 *   Create a low priority thread which will only become active when
 *   all other threads have suspended.  This thread should:
 *	1) Call _p0_blocking_poll(), to check for available messages.
 *	2) Keep track of idle time.
 *
 *   If a thread module does not support priorities, this same
 *   functionality can be hacked into ports0_cond_wait() using a
 *   counter of suspend threads.
 *
 * TODO: We can probably just use ports0_thread_create() once attributes are
 * implemented.
 *
 **  I just used ports0_thread_create since it seems to do everything required.
 **  If the global handler thread cannot be identical to the idle thread, then
 **  a general priority scheme has to be implemented.
 */
void _p0_thread_create_handler_thread()
{
    if (!_p0_must_poll())
    {
	ports0_debug_printf(2, ("_p0_thread_create_handler_thread(): Not creating handler thread\n"));
	return;
    }
    
    ports0_debug_printf(2,("_p0_thread_create_handler_thread(): Creating idle thread\n"));

    using_handler_thread = PORTS0_TRUE;
    
    ports0_thread_create( &handler_thread, NULL, handler_thread_func, NULL );

    pP.idleThread = handler_thread;
    
} /* _p0_thread_create_handler_thread() */


/*
 * handler_thread_func()
 *
 * TODO: Should count 'unblocked_threads' as they call and return
 * from cond_wait.  Thus, when all threads except this one are blocked,
 * we know we can call _p0_blocking_poll() instead of ports0_poll when
 * using preemptive threads.
 */
static void *handler_thread_func(void *arg)
{
  int im_done;

    ports0_debug_printf(2,("handler_thread_func(): Entering\n"));

    for  (;;)  {
      ports0_debug_printf(2,("handler_thread_func(): Has been scheduled\n"));

      ports0_mutex_lock( &handler_mutex );
      im_done = handler_thread_done;
      ports0_mutex_unlock( &handler_mutex );
      if  ( im_done )  break;

      if ( preemptive_threads  || (pP.lastLook->qhead != NULL) )
	{
	  ports0_poll();
	}
      else
	{
	  /* start_idle_timer(); */
	  _p0_blocking_poll();
	  /* stop_idle_timer();  */
	}
      
      ports0_debug_printf(2,("handler_thread_func(): calling yield\n"));
      ports0_thread_yield();
    }

    ports0_debug_printf(2,("handler_thread_func(): Exiting\n"));

    ports0_cond_signal( &handler_cond );
    ports0_thread_exit( ports0_thread_self() );
    return (NULL);
} /* handler_thread_func() */


/*
 * _p0_thread_shutdown_handler_thread()
 * This in effect does a join on the handler thread _unless_ this message is being
 * handled as a result of the handler thread doing a poll, that is to say, the
 * handler thread is trying to shut itself down.
 */
void _p0_thread_shutdown_handler_thread()
{
    if (using_handler_thread)
    {
      ports0_mutex_lock( &handler_mutex );
      handler_thread_done = PORTS0_TRUE;
      if  ( ports0_thread_self() != pP.idleThread )
	ports0_cond_wait( &handler_cond, &handler_mutex );
      ports0_mutex_unlock( &handler_mutex );
    }
} /* _p0_thread_shutdown_handler_thread() */


/*
 * ports0_preemptive_threads
 *
 * Return PORTS0_TRUE (non-zero) if we are using preemptive threads.
 */
bool_t ports0_preemptive_threads()
{
    return (preemptive_threads);
} /* ports0_preemptive_threads() */


/*
 * ports0_thread_key_create()
 */
int ports0_thread_key_create( ports0_thread_key_t *key, ports0_thread_key_destructor_func_t dfunc )
{
  int i;

  DONT_PREEMPT;
  spin_lock( &key_lock );
  if  ( used_keys == PORTS0_DATAKEYS_MAX )  {
    spin_unlock( &key_lock );
    errno = EAGAIN;
    return -1;
  }
  i = used_keys++;

  pP.activeThread->key_data[i] = NULL;
  key_destructor[i] = dfunc;
  spin_unlock( &key_lock );
  CHECK_SLICE;

  *key = i;
  return 0;
}  /* ports0_thread_key_create() */


/*
 * ports0_thread_setspecific()
 */
int ports0_thread_setspecific( ports0_thread_key_t key, void *value )
{
  if  ( key < 0  || used_keys <= key )
    { errno = EINVAL;  return -1;  }

  pP.activeThread->key_data[key] = value;
  return 0;
}


/*
 * ports0_thread_getspecific()
 */
int ports0_thread_getspecific( ports0_thread_key_t key, void **value )
{
  if  ( key < 0  || used_keys <= key )
    { errno = EINVAL;  return -1;  }

  *value = pP.activeThread->key_data[key];
  return 0;
}


/*
 *  ports0_thread_once()
 */
int ports0_thread_once( ports0_thread_once_t *once_cntl, void (*init_routine)() )
{
  DONT_PREEMPT;
  spin_lock( &(once_cntl->lock) );
  if  ( once_cntl->init )  {
    once_cntl->init = 0;
    spin_unlock( &(once_cntl->lock) );
    (*init_routine)();
  }  else  spin_unlock( &(once_cntl->lock) );
  CHECK_SLICE;
  return 0;
}


void ports0_thread_yield()
{
  ports0_thread_t self = pP.activeThread;

  DONT_PREEMPT;
  if  ( sched_next_activethread(NON_BLOCKING_SCHED) )  {
    ReadyQPut_switch( self, pP.activeThread );
    /* Don't need to CHECK_SLICE here since we just did a csw. */
  }  else  {
    pP.activeThread = self;
    START_PREEMPTION;
    /* We can just throw away timer hits since there's no one to csw to. */
  }
}


void ports0_thread_exit( void *xstatus )
{
  cleanup( pP.activeThread );
}

/*
 *
 *  Mutex stuff
 *
 */

int _p0_mutex_init( ports0_mutex_t *nmutex, ports0_mutexattr_t *attr )
{
  nmutex->mutex.locked = MUTEX_UNLOCKED;
  ThreadQInit( &(nmutex->mutex.waiters) );
  return 0;
}


int _p0_mutex_destroy( ports0_mutex_t *nmutex )
{
  nmutex->mutex.locked = MUTEX_UNLOCKED;
  ThreadQInit( &(nmutex->mutex.waiters) );
  return 0;
}


int _p0_mutex_lock( ports0_mutex_t *nmutex )
{
  ports0_thread_t self;

  DONT_PREEMPT;
  spin_lock( &(nmutex->waiters.lock) );

  if  ( nmutex->mutex.locked )  {
    self = pP.activeThread;
    Ports0Assert2(( (nmutex->mutex.locked >> 1) != self->tid),
		 ("Recursive mutex error: tid = %d\n", self->tid) );
    ThreadPut( self, &(nmutex->mutex.waiters) );
    sched_next_activethread(BLOCKING_SCHED);
#ifdef MULTIPROC
    QT_BLOCK( csw_helper, self, &(nmutex->mutex.waiters.lock), pP.activeThread->sp );
#else
    QT_BLOCK( csw_helper, self, NULL, pP.activeThread->sp );
#endif

    /*  This thread is resumed with MUTEX_LOCKED already.       */
    /*  Don't need to CHECK_SLICE here since we just did a csw. */

  }  else  {

    /*  This simply allows a debugger to see who's holding a mutex. */
    nmutex->mutex.locked = (pP.activeThread->tid << 1) | MUTEX_LOCKED;
    spin_unlock( &(nmutex->mutex.waiters.lock) );
    CHECK_SLICE;

  }

  return 0;
}


int _p0_mutex_unlock( ports0_mutex_t *nmutex )
{
  ports0_thread_t head_waiter;

  DONT_PREEMPT;
  spin_lock( &(nmutex->mutex.waiters.lock) );

  if  ( nmutex->mutex.waiters.qhead )  {

    ThreadGet( head_waiter, &(nmutex->mutex.waiters) );
    nmutex->mutex.locked = (head_waiter->tid << 1) | MUTEX_LOCKED;
    spin_unlock( &(nmutex->mutex.waiters.lock) );
    /*  Leave MUTEX_LOCKED and resume this guy.  */
    ReadyQPut( head_waiter );

  }  else  {

    nmutex->mutex.locked = MUTEX_UNLOCKED;
    spin_unlock( &(nmutex->mutex.waiters.lock) );

  }

  CHECK_SLICE;
  return 0;
}


/*
 *
 *  Condition variable stuff
 *
 */
int _p0_cond_init( ports0_cond_t *ncond, ports0_condattr_t *attr )
{
  ncond->cond.mutex = NULL;
  ThreadQInit( &(ncond->cond.waiters) );
  return 0;
}


int _p0_cond_destroy( ports0_cond_t *ncond )
{
  ncond->cond.mutex = NULL;
  ThreadQInit( &(ncond->cond.waiters) );
  return 0;
}


int _p0_cond_wait( ports0_cond_t *ncond, ports0_mutex_t *mutex )
{
  ports0_thread_t self;

  DONT_PREEMPT;
  /* Don't need to CHECK_SLICE later since we're headed for a csw anyway. */
  spin_lock( &(ncond->cond.waiters.lock) );

  /*  Put myself on the condition waiters list, release the mutex,
      and schedule another thread.
  */

  self = pP.activeThread;
  ThreadPut( self, &(ncond->cond.waiters) );
  ports0_mutex_unlock( mutex );
  sched_next_activethread(BLOCKING_SCHED);
#ifdef MULTIPROC
  QT_BLOCK( csw_helper, self, &(ncond->cond.waiters.lock), pP.activeThread->sp );
#else
  QT_BLOCK( csw_helper, self, NULL, pP.activeThread->sp );
#endif

  /*  When I come back here, I need to have the mutex.  Since the signaler
	need not own the mutex lock, I can't be guarenteed of being resumed
	from the mutex waiter list with the mutex lock.  I've got to contend
	for it from my ReadyQ like everybody else.
  */
  ports0_mutex_lock( mutex );

  return 0;
}


int _p0_cond_signal( ports0_cond_t *ncond )
{
  ports0_thread_t head_waiter;

  DONT_PREEMPT;
  spin_lock( &(ncond->cond.waiters.lock) );

  /*  Resume head waiter (if one), unlock and return.  */

  if  ( ncond->cond.waiters.qhead )  {
    ThreadGet( head_waiter, &(ncond->cond.waiters) );
    spin_unlock( &(ncond->cond.waiters.lock) );
    ReadyQPut( head_waiter );
  }  else  {
    spin_unlock( &(ncond->cond.waiters.lock) );
  }    

  CHECK_SLICE;
  return 0;
}


int _p0_cond_broadcast( ports0_cond_t *ncond )
{
  DONT_PREEMPT;
  spin_lock( &(ncond->cond.waiters.lock) );

  /*  Resume all waiters (if any), unlock and return.  */

  if  ( ncond->cond.waiters.qhead )  {
    ReadyQCat( ncond->cond.waiters.qhead, ncond->cond.waiters.qtail );
    ThreadQInit( &(ncond->cond.waiters) );
  }  else  {
    spin_unlock( &(ncond->cond.waiters.lock) );
  }    

  CHECK_SLICE;
  return 0;
}


int get_activethreads()
{ return TKp->ActiveThreads; }

int get_mytid()
{ return pP.activeThread->tid; }


/*
 * _p0_thread_shutdown()
 */
void _p0_thread_shutdown()
{
} /* _p0_thread_shutdown() */


/*
 * _p0_thread_abort()
 */
void _p0_thread_abort(int return_code)
{
} /* _p0_thread_abort() */


void _p0_report_bad_rc( int rc, char *message )
{
  char achMessHead[] = "[Thread System]";

  if( rc!=_P0_THR_SUCCESS ) {
    ports0_stdio_lock();
    fprintf( stderr, message );
    ports0_stdio_unlock();
    DONT_PREEMPT;
    switch( errno ) {
    case EAGAIN:
      ports0_fatal("%s %s\n%s system out of resources (EAGAIN)\n",
		  achMessHead, message, achMessHead);
      break;
    case ENOMEM:
      ports0_fatal("%s %s\n%s insufficient memory (ENOMEM)\n",
		  achMessHead, message, achMessHead );
      break;
    case EINVAL:
      ports0_fatal("%s %s\n%s invalid value passed to thread interface (EINVAL)\n",
		  achMessHead, message, achMessHead );
      break;
    case EPERM:
      ports0_fatal("%s %s\n%s user does not have adequate permission (EPERM)\n",
		  achMessHead, message, achMessHead );
      break;
    case ERANGE:
      ports0_fatal("%s %s\n%s a parameter has an invalid value (ERANGE)\n",
		  achMessHead, message, achMessHead );
      break;
    case EBUSY:
      ports0_fatal("%s %s\n%s mutex is locked (EBUSY)\n",
		  achMessHead, message, achMessHead );
      break;
    case EDEADLK:
      ports0_fa  Nexus.create is Noelle's 
draft.  Nexus2.create is the same thing, with my manipulation on 2 elements.
tal("%s %s\n%s deadlock detected (EDEADLK)\n",
		  achMessHead, message, achMessHead );
      break;
    case ESRCH:
      ports0_fatal("%s %s\n%s could not find specified thread (ESRCH)\n",
		  achMessHead, message, achMessHead );
      break;
    default:
      ports0_fatal("%s %s\n%s unknown error number: %d\n",
		  achMessHead, message, achMessHead, errno );
      break;
    }
    _p0_imprison_thread( "Error code detected\n" );
/*    _p0_exit_process( 1 ); */
  }
}
