/*
 * th_qt.h
 *
 * author: Craig Lee at Aerospace.
 *
 * External header for QuickThreads running on the
 * whatever QuickThreads runs on
 *
 * rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_qt.h,v 1.1 1994/11/01 03:53:45 tuecke Exp $"
 */

#include <qt.h>

#ifndef PAGESIZE
#ifdef PORTS0_ARCH_OSF1
#define PAGESIZE (sysconf(_SC_PAGESIZE))
#endif
#endif

#define DEFAULT_STACK_PAGES 10
#define _PORTS0_THREAD_ATTR_STACKSIZE
#define PORTS0_STACK_MIN  DEFAULT_STACK_PAGES*PAGESIZE

#define NON_BLOCKING_SCHED 0
#define BLOCKING_SCHED 1

/* Time quanta in usec: clk_res <= TIME_SLICE < 10^6 */
#define TIME_SLICE 100000
#define START_PREEMPTION { \
			   timer_hit = 0; \
			   dont_preempt = 0; \
			 }
#define DONT_PREEMPT dont_preempt = 1
#define CHECK_SLICE { \
		      if  ( timer_hit )  { \
		        check_slice_hits++; \
			ports0_thread_yield(); \
		      } \
		      dont_preempt = 0; \
		    }

#define PORTS0_QT
#define _P0_THR_SUCCESS 0

#define _P0_THREAD_GRAN 256
#define _P0_USER_THREAD 0

/*
 * lock stuff
 */

#ifdef MULTIPROC

#define LOCK_INIT_VAL 0
typedef  int  spin_lock_t;

void spin_lock_init( spin_lock_t *lock );
void spin_lock( spin_lock_t *lock );
int  test_and_get_spin_lock( spin_lock_t *lock );  /* not currently used */
void spin_unlock( spin_lock_t *lock );

#else

#define spin_lock_init(lp)
#define spin_lock(lp)
#define test_and_get_spin_lock(lp)
#define spin_unlock(lp)

#endif

/*
 * ports0_thread_t stuff
 */

#define PORTS0_DATAKEYS_MAX 16
typedef void (*ports0_thread_key_destructor_func_t)(void *value);
typedef int ports0_thread_key_t;

typedef struct ports0_thread_struct {
    int tid;
    int stack_size;
    char *stack_base;
    qt_t *sp;
    void *key_data[PORTS0_DATAKEYS_MAX];
    struct ports0_thread_struct *next;
}  ports0_thread_s;

typedef  ports0_thread_s * ports0_thread_t;

typedef struct ports0_thread_attr_s {
  int stack_addr;
  int stack_size;
}  ports0_thread_attr_t;


typedef struct thread_queue {
#ifdef MULTIPROC
  spin_lock_t lock;
#endif
  ports0_thread_t qhead;
  ports0_thread_t qtail;
}  ThreadQ;


/*
 * ports0_mutex_t stuff
 */

#define  MUTEX_UNLOCKED  0;
#define  MUTEX_LOCKED    1;

typedef struct ports0_mutex_s {
  int locked;
  ThreadQ waiters;
}  p0_mutex;

typedef struct ports0_mutexattr_s {
  int dummy;
}  ports0_mutexattr_t;

/*
 * ports0_mutex_t
 */
typedef struct _ports0_mutex_t {
  _P0_START_MAGIC_COOKIE
  p0_mutex mutex;
  _P0_END_MAGIC_COOKIE
} ports0_mutex_t;

/*
 * ports0_cond_t stuff
 */

typedef struct ports0_cond_s {
  ports0_mutex_t *mutex;
  ThreadQ waiters;
}  p0_cond;

typedef struct ports0_condattr_s {
  int dummy;
}  ports0_condattr_t;

 
/*
 * ports0_cond_t
 */
typedef struct _ports0_cond_t {
  _P0_START_MAGIC_COOKIE
  p0_cond cond;
  _P0_END_MAGIC_COOKIE
} ports0_cond_t;

typedef struct _ports0_global_vars {
  ports0_thread_key_t        ports0_thread_t_pointer;
} ports0_global_vars;
 
PORTS0_GLOBAL ports0_global_vars ports0_all_global_vars;

/*
 * Thread Kernel stuff
 */

#define  MAXPROCS 128

  /* Global information */
typedef struct thread_kernel_s {
  int IdleProcs;
  int idle_vec[MAXPROCS];
  int ActiveThreads;
  int nThreads;
  ThreadQ *firstQ, *lastQ;
  ThreadQ readyList[MAXPROCS];
  int free_count;
  ThreadQ freeQ;
}  thread_kernel_t;


  /* Process private information */
typedef struct private_proc {
  int myId;
  char *signal_stack;
  ports0_thread_t activeThread;
  ports0_thread_t idleThread;
  ThreadQ *lastLook;
  ThreadQ *putQ;
  /* A per process free list could be implemented here. */
}  PrivateProcessor;


#ifdef MULTIPROC
/*  In shared memory, these functions have to
    be mapped to locking thread-safe versions.
    Perhaps these should all just be ports0_malloc calls.
*/
#else
#define nq_minit(X)
#define nq_malloc(X) malloc(X)

#ifdef PORTS0_ARCH_OSF1
#define nq_memalign(X,Y) valloc(Y)
#else
#define nq_memalign(X,Y) memalign(X,Y)
#endif

#define nq_valloc(X) valloc(X)
#define nq_free(X) free(X)

#define atomic_inc(x) (++(x))
#define atomic_dec(x) (--(x))
#endif


#ifdef MULTIPROC
#define PORTS0_THREAD_ONCE_INIT {LOCK_INIT_VAL,1};
typedef struct ports0_thread_once_s {
  spin_lock_t lock;
  int init;
}  ports0_thread_once_t;
#else
#define PORTS0_THREAD_ONCE_INIT {1};
typedef struct ports0_thread_once_s {
  int init;
}  ports0_thread_once_t;
#endif


#define NextQ(q) ((q == TKp->lastQ) ? TKp->firstQ : (q + 1))

#define ThreadQInit(t) {\
	spin_lock_init( &((t)->lock) ); \
	(t)->qhead = (t)->qtail = NULL; \
        }

#define ThreadPut(t,q) { \
	if  ( (q)->qhead ) \
          { (q)->qtail->next = t;  (q)->qtail = t; } \
        else \
          { (q)->qhead = (q)->qtail = t; } \
        }

#define ThreadCat(hd,tl,q) { \
	if  ( q->qhead )  { q->qtail->next = hd; } \
        else              { q->qhead = hd; } \
        q->qtail = tl; \
        }

#define ThreadGet(t,q) { \
        (t) = (q)->qhead; \
	(q)->qhead = (t)->next; \
        (t)->next = NULL; \
        }

#define ReadyQPut(t) { \
	spin_lock( &(pP.putQ->lock) ); \
	ThreadPut( t, (pP.putQ) ); \
	spin_unlock( &(pP.putQ->lock) ); \
	}

#define ReadyQCat(hd,tl) { \
	spin_lock( &(pP.putQ->lock) ); \
	ThreadCat( hd, tl, (pP.putQ) ); \
	spin_unlock( &(pP.putQ->lock) ); \
	}

#define FreeQPut(t) { \
	spin_lock( &(TKp->freeQ.lock) ); \
	ThreadPut( t, (&(TKp->freeQ)) ); \
	(TKp->free_count)++; \
	spin_unlock( &(TKp->freeQ.lock) ); \
	}

#define FreeQGet(t,ss) { \
	spin_lock( &(TKp->freeQ.lock) ); \
	if  ( (t) = TKp->freeQ.qhead ) { \
          if  ( (ss) <= (t)->stack_size )  { \
	    TKp->freeQ.qhead = (t)->next; \
	    (TKp->free_count)--; \
          }  else \
            { (t) = NULL; } \
        } \
	spin_unlock( &(TKp->freeQ.lock) ); \
	}

/*  There are two versions of these two macros since in the multiprocessor
    version, the queue lock pointer has to be passed to csw_helper to
    unlock the queue after the calling thread has been safely put away.
*/
#ifdef MULTIPROC

#define ReadyQPut_switch(t,n) { \
	spin_lock( &(pP.putQ->lock) ); \
	ThreadPut( t, (pP.putQ) ); \
	QT_BLOCK( csw_helper, (t), &(pP.putQ->lock), (n)->sp ); \
	}

#define FreeQPut_switch(t,n) { \
	spin_lock( &(TKp->freeQ.lock) ); \
	ThreadPut( t, (&(TKp->freeQ)) ); \
	(TKp->free_count)++; \
	QT_ABORT( csw_helper, (t), &(TKp->freeQ.lock), (n)->sp ); \
	}

#else

#define ReadyQPut_switch(t,n) { \
	ThreadPut( t, (pP.putQ) ); \
	QT_BLOCK( csw_helper, (t), NULL, (n)->sp ); \
	}

#define FreeQPut_switch(t,n) { \
	ThreadPut( t, (&(TKp->freeQ)) ); \
	(TKp->free_count)++; \
	QT_ABORT( csw_helper, (t), NULL, (n)->sp ); \
	}

#endif  /* MULTIPROC */


#ifdef __cplusplus
extern "C" {
#endif

extern void _p0_more_thread_mem( void );
extern void _p0_init_thread_package( void );
extern void *_p0_thread_starter( void * );
extern void _p0_report_bad_rc( int, char * );

#ifdef __cplusplus
}
#endif

#define _p0_samethread( t1, t2 ) ( ports0_thread_equal( t1, t2 ) )

#define _p0_current_thread( t1 ) (t1)=ports0_thread_self()

#define _p0_test_rc( a, b ) \
  do {                                                   \
    if( a != _P0_THR_SUCCESS ) _p0_report_bad_rc( a, b ); \
  } while(0)


extern int   ports0_thread_create(ports0_thread_t *thread,
				   ports0_thread_attr_t *attr,
				   ports0_thread_func_t func,
				   void *user_arg);
extern void  ports0_thread_exit(void *status);

extern int  ports0_thread_key_create(ports0_thread_key_t *key,
				    ports0_thread_key_destructor_func_t func);
extern int  ports0_thread_setspecific(ports0_thread_key_t key,
				     void *value);
extern int ports0_thread_getspecific(ports0_thread_key_t key,
				    void **value);
extern ports0_thread_t ports0_thread_self( void );

extern void ports0_thread_yield( void );

extern int _p0_mutex_init( ports0_mutex_t *, ports0_mutexattr_t * );
extern int _p0_mutex_destroy( ports0_mutex_t * );
extern int _p0_cond_init( ports0_cond_t *, ports0_condattr_t *);
extern int _p0_cond_destroy( ports0_cond_t *);
extern int _p0_mutex_lock( ports0_mutex_t * );
extern int _p0_mutex_unlock( ports0_mutex_t * );
extern int _p0_cond_wait( ports0_cond_t *, ports0_mutex_t * );
extern int _p0_cond_signal( ports0_cond_t * );
extern int _p0_cond_broadcast( ports0_cond_t * );


#ifdef PORTS0_SKIP_USE_MACROS
#undef PORTS0_USE_MACROS
#endif

#ifdef PORTS0_USE_MACROS

#define ports0_thread_equal(t1,t2) ((t1) == (t2))

#define ports0_mutex_init( mut, attr ) \
  do {                                                      \
     int rc;                                                \
                                                            \
     _P0_INIT_START_MAGIC_COOKIE(mut);                      \
     rc = _p0_mutex_init( mut, attr );          \
     _p0_test_rc( rc, "PORTS0: allocate lock failed\n" );    \
     _P0_INIT_END_MAGIC_COOKIE(mut);                        \
      } while(0)

#define ports0_mutex_destroy( mut )                           \
 do {                                                        \
   int rc;                                                   \
                                                             \
   PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "destroy" );         \
   rc = _p0_mutex_destroy( mut );                 \
   _p0_test_rc( rc, "PORTS0: free lock failed\n" );           \
 } while(0)

#define ports0_cond_init( cv, attr )                           \
 do {                                                        \
   int rc;                                                   \
                                                             \
   _P0_INIT_START_MAGIC_COOKIE(cv);                          \
   rc = _p0_cond_init( cv, attr );                 \
   _p0_test_rc( rc, "PORTS0: allocate condition variable failed\n"); \
   _P0_INIT_END_MAGIC_COOKIE(cv);                            \
 } while(0)

#define ports0_cond_destroy( cv )                           \
 do {                                                            \
   int rc;                                                       \
                                                                 \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "destroy" );               \
   rc = _p0_cond_destroy( cv );                  \
   _p0_test_rc( rc, "PORTS0: free condition variable failed\n" ); \
 } while(0)

#define ports0_mutex_lock( mut )                                     \
 do {                                                          \
  int rc;                                                      \
                                                               \
  PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock1" );              \
  rc = _p0_mutex_lock( mut );                            \
  _p0_test_rc( rc, "PORTS0: mutex lock failed\n" );             \
  PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock2" );              \
 } while(0)

#define ports0_mutex_unlock( mut )                                   \
 do {                                                          \
   int rc;                                                     \
                                                               \
   PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "unlock1" );           \
   rc = _p0_mutex_unlock( mut );                         \
   _p0_test_rc( rc, "PORTS0: mutex unlock failed\n" );          \
 } while(0)

#define ports0_cond_wait( cv, mut )                            \
 do {                                                               \
   int rc;                                                          \
                                                                    \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "wait1" );                    \
   rc = _p0_cond_wait( cv, mut );       \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "wait2" );                    \
   _p0_test_rc( rc, "PORTS0: condition variable wait failed\n" );    \
 } while(0)

#define ports0_cond_signal( cv )                               \
 do {                                                               \
   int rc;                                                          \
                                                                    \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "signal1" );                  \
   rc = _p0_cond_signal( cv );                      \
   _p0_test_rc( rc, "PORTS0: condition variable signal failed\n" );  \
 } while(0)

#define ports0_cond_broadcast( cv )                            \
 do {                                                               \
   int rc;                                                          \
                                                                    \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "signal1" );                  \
   rc = _p0_cond_broadcast( cv );                   \
   _p0_test_rc( rc, "PORTS0: condition variable signal failed\n" );  \
 } while(0)

#else

extern int ports0_thread_equal(ports0_thread_t t1, ports0_thread_t t2);

#define ports0_mutex_init( mut, attr ) _p0_mutex_init( mut, attr )
#define ports0_mutex_destroy( mut )    _p0_mutex_destroy( mut )
#define ports0_cond_init( cv, attr )   _p0_cond_init( cv, attr )
#define ports0_cond_destroy( cv )      _p0_cond_destroy( cv )
#define ports0_mutex_lock( mut )       _p0_mutex_lock( mut )
#define ports0_mutex_unlock( mut )     _p0_mutex_unlock( mut )
#define ports0_cond_wait( cv, mut )    _p0_cond_wait( cv, mut )
#define ports0_cond_signal( cv )       _p0_cond_signal( cv )
#define ports0_cond_broadcast( cv )    _p0_cond_broadcast( cv )

#endif


/* Gets ports0 wrapper thread id from pthread specific data */

typedef void *p0_any;

#define _p0_thread_self(Thread) \
 do {  \
  int rc;  \
  rc = ports0_thread_getspecific( ports0_all_global_vars.ports0_thread_t_pointer, \
			      (p0_any *)(Thread)); \
  _p0_test_rc( rc, "PORTS0: get thread-local data failed\n"); \
 } while(0)
