
static char *rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_sol.c,v 1.1 1994/11/01 03:53:58 tuecke Exp $";

#include "internal.h"

#define _P0_THREAD_GRAN 256
#define _P0_USER_THREAD 0

static bool_t	preemptive_threads = PORTS0_TRUE;
static bool_t	arg_got_sched;
static int 	scheduler;
static int	priority_min;
static int	priority_max;
static int	priority_mid;
static int      arg_concurrency_level;

static bool_t		using_handler_thread = PORTS0_FALSE;
static unsigned int	handler_thread;
static bool_t		handler_thread_done = PORTS0_FALSE;
static void *		handler_thread_func(void *arg);

/*
 * Free list of _p0_thread_t structures
 */
static _p0_thread_t *	Thread_Freelist;
static ports0_mutex_t	thread_mem_mutex;
static int              next_thread_id;
static int		number_of_threads;

static void *           thread_starter(void *temparg);
static _p0_thread_t *   new_thread(void);
static void             set_tsd(_p0_thread_t *);

/*
 * For ports0_thread_once()
 */
static ports0_mutex_t	thread_once_mutex;

/*
 * _p0_thread_process_arguments_init()
 *
 * Initialize any variables that will be used to hold
 * my configuration information extracted from arguments.
 */
void _p0_thread_process_arguments_init()
{
  arg_concurrency_level = 5;
} /* _p0_thread_process_arguments_init() */


/*
 * _p0_thread_process_arguments()
 *
 * If ports0_package_getarg(current_arg) is destined for my module,
 * then extract it into my own variables.
 *
 * Return:	A new current argument.
 */
int _p0_thread_process_arguments(int current_arg, int arg_count)
{
  char *arg;

  arg = ports0_package_getarg(current_arg);
  if ((strcmp(arg, "-concurrency_level") == 0) &&
      ((current_arg + 1) < arg_count))
    {
      arg_concurrency_level = atoi(ports0_package_getarg(current_arg + 1));
      current_arg += 2;
    }

    return (current_arg);
} /* _p0_thread_process_arguments() */


/*
 * _p0_thread_usage_message()
 *
 * Print a usage message for this module's arguments to stdout.
 */
void _p0_thread_usage_message()
{
    printf("    -concurrency_level <integer> : Solaris concurrency level,\n");
} /* _p0_thread_usage_message() */


/*
 * _p0_thread_new_process_params()
 *
 * Write any new process parameters for this module into 'buf',
 * up to a maximum of 'size' characters.
 *
 * Return:	The number of characters written into 'buf'.
 */
int _p0_thread_new_process_params(char *buf, int size)
{
    char tmp_buf1[1024];
    char tmp_buf2[1024];
    int n_added;

    tmp_buf1[0] = '\0';

    sprintf(tmp_buf1, "-concurrency_level %d ",
	    arg_concurrency_level);
    n_added = strlen(tmp_buf1);
    if (n_added > size)
	ports0_fatal("_p0_thread_new_process_params(): Internal error: Not enough room in buffer for arguments\n");

    strcpy(buf, tmp_buf1);
    
    return (n_added);
} /* _p0_thread_new_process_params() */


/*
 * _p0_thread_preinit()
 *
 * If you need to call a thread package initialization routine, then
 * do it here.  This is called as the very first thing in ports0_init().
 */
void _p0_thread_preinit( void )
{
} /* _p0_thread_preinit() */


void _p0_thread_init( void )
{
  int i;
  int rc=0;
  _p0_thread_t  *thread;

  preemptive_threads = PORTS0_TRUE;
  while((rc = thr_setconcurrency( arg_concurrency_level ))==4) ;
  _p0_test_rc( rc, "PORTS0: thr_setconcurrency failed\n");
  while((rc = thr_keycreate( &(ports0_all_global_vars.ports0_thread_t_pointer), NULL))==4) ;
  _p0_test_rc( rc, "PORTS0: thr_keycreate failed\n");
  (ports0_all_global_vars.general_attribute) = USYNC_THREAD;
  (ports0_all_global_vars.thread_flags) = THR_DETACHED;

  ports0_mutex_init( &(thread_mem_mutex), (ports0_mutexattr_t *) NULL );

  /* Initialize the thread_once_mutex */
  ports0_mutex_init(&thread_once_mutex, (ports0_mutexattr_t *) NULL);
  next_thread_id = 0;
  
  /* Initialize the _p0_thread_t structure for this initial thread */
  thread = new_thread();
  thread->names = (ports0_thread_name_list_t *) NULL;
  set_tsd( thread );
}

/*
 * _p0_thread_shutdown()
 */
void _p0_thread_shutdown()
{
} /* _p0_thread_shutdown() */


/*
 * _p0_thread_abort()
 */
void _p0_thread_abort(int return_code)
{
} /* _p0_thread_abort() */

void _p0_report_bad_rc( int rc, char *message )
{
  char achMessHead[] = "[Thread System]";
  if( rc!=_P0_THR_SUCCESS ) {
    fprintf( stderr, message );
    switch( rc ) {
    case EAGAIN:
      ports0_fatal("%s %s\n%s system out of resources (EAGAIN)\n",
		  achMessHead, message, achMessHead);
      break;
    case ENOMEM:
      ports0_fatal("%s %s\n%s insufficient memory (ENOMEM)\n",
		  achMessHead, message, achMessHead );
      break;
    case EINVAL:
      ports0_fatal("%s %s\n%s invalid value passed to thread interface (EINVAL)\n",
		  achMessHead, message, achMessHead );
      break;
    case EPERM:
      ports0_fatal("%s %s\n%s user does not have adequate permission (EPERM)\n",
		  achMessHead, message, achMessHead );
      break;
    case ERANGE:
      ports0_fatal("%s %s\n%s a parameter has an invalid value (ERANGE)\n",
		  achMessHead, message, achMessHead );
      break;
    case EBUSY:
      ports0_fatal("%s %s\n%s mutex is locked (EBUSY)\n",
		  achMessHead, message, achMessHead );
      break;
    case EDEADLK:
      ports0_fatal("%s %s\n%s deadlock detected (EDEADLK)\n",
		  achMessHead, message, achMessHead );
      break;
    case ESRCH:
      ports0_fatal("%s %s\n%s could not find specified thread (ESRCH)\n",
		  achMessHead, message, achMessHead );
      break;
    default:
      ports0_fatal("%s %s\n%s unknown error number: %d\n",
		  achMessHead, message, achMessHead, rc );
      break;
    }
    _p0_imprison_thread( "Error code detected\n" );
/*    _p0_exit_process( 1 ); */
  }
}

static _p0_thread_t *new_thread( void )
{
  int i;
  _p0_thread_t *new_thread;
  int mem_req_size;
 
  ports0_mutex_lock(&thread_mem_mutex);

  if( Thread_Freelist==NULL ) {
    mem_req_size = sizeof(_p0_thread_t) * _P0_THREAD_GRAN;
    Ports0Malloc(_p0_more_thread_mem(), Thread_Freelist, _p0_thread_t *, 
		mem_req_size);

    for( i=0; i<_P0_THREAD_GRAN-1; i++ ) {
      Thread_Freelist[i].next_free=&Thread_Freelist[i+1];
    }
    Thread_Freelist[_P0_THREAD_GRAN-1].next_free=NULL;
  }
  new_thread = Thread_Freelist;
  if( Thread_Freelist!=NULL ) Thread_Freelist=Thread_Freelist->next_free;

  new_thread->id = next_thread_id++;

  ports0_mutex_unlock(&thread_mem_mutex);

  return new_thread;

}

static void terminate_thread(_p0_thread_t *thread, void *status)
{
    int i;
#ifdef PORTS0_PROFILE
    int node_id, context_id;
#endif
    

#ifdef PORTS0_PROFILE
    _p0_node_id(&node_id);
    _p0_context_id(&context_id);
    _p0_pablo_log_thread_destruction(node_id, context_id, thread->id);
#endif /* PORTS0_PROFILE */    

    /* Free up the thread storage */
    ports0_mutex_lock(&thread_mem_mutex);
    thread->next_free = Thread_Freelist;
    Thread_Freelist = thread;
    ports0_mutex_unlock(&thread_mem_mutex);

    /* Exit the thread */
    thr_exit(NULL);

} /* terminate_thread() */


void ports0_thread_exit( void *status )
{
  _p0_thread_t *victim;
  
  _p0_thread_self(&victim);

  terminate_thread(victim, status);
}

static void set_tsd( _p0_thread_t * thread )
{
  int rc=0;

  while((rc=thr_setspecific( ports0_all_global_vars.ports0_thread_t_pointer, (void *) thread ))==4) ;
  _p0_test_rc( rc, "PORTS0: set thread-local data failed\n" );
}

static void *thread_starter( void *temparg )
{
  _p0_thread_t *thread;
  int i;
  void *status;
#ifdef PORTS0_PROFILE
  int node_id, context_id;
#endif

  thread = (_p0_thread_t *)temparg;

  set_tsd( thread );

  /* Set the context of this thread */
  ports0_thread_setspecific(_p0_context_key, thread->context);
    
#ifdef PORTS0_PROFILE
    _p0_node_id(&node_id);
    _p0_context_id(&context_id);
    _p0_pablo_log_thread_creation(node_id, context_id, thread->id);
#endif /* PORTS0_PROFILE */    
    
  /* Call the user function */
  status = (*thread->user_func)(thread->user_arg);
  
  /* Terminate the thread */
  terminate_thread(thread, status);
  
  return (NULL);
}

/*
 * ports0_thread_create
 */

int ports0_thread_create(ports0_thread_t *user_thread,
			ports0_thread_attr_t *attr,
			ports0_thread_func_t func,
			void *user_arg )
{
  _p0_thread_t *thread;
  unsigned int thread_id;
  int rc=0;


  thread = new_thread();
  thread->names = (ports0_thread_name_list_t *) NULL;
  
  /* Initialize the thread data that needs to be passed to the new thread */
  ports0_thread_getspecific(_p0_context_key, &(thread->context));
  thread->user_func = func;
  thread->user_arg = user_arg;

  while((rc = thr_create( NULL, 0, thread_starter,
			  thread, (ports0_all_global_vars.thread_flags),
			 &thread_id ))==4);
  _p0_test_rc( rc, "PORTS0: create thread failed\n" );

  if(user_thread) {
    *user_thread = thread->id;
  }

  return (0);
}

/*
 * _p0_thread_create_handler_thread()
 *
 * If we have preemptive threads:
 *   Create a normal priority thread that will sit in a
 *   loop calling ports0_poll().
 *
 * If we have non-preemptive threads:
 *   Create a low priority thread which will only become active when
 *   all other threads have suspended.  This thread should:
 *	1) Call _p0_blocking_poll(), to check for available messages.
 *	2) Keep track of idle time.
 *
 *   If a thread module does not support priorities, this same
 *   functionality can be hacked into ports0_cond_wait() using a
 *   counter of suspend threads.
 *
 * TODO: We can probably just use ports0_thread_create() once attributes are
 * implemented.
 */
void _p0_thread_create_handler_thread()
{
    int rc;
    _p0_thread_t *thread;
    int thread_id;

    if (!_p0_must_poll())
    {
	ports0_debug_printf(2, ("_p0_thread_create_handler_thread(): Not creating handler thread\n"));
	return;
    }

    ports0_debug_printf(2, ("_p0_thread_create_handler_thread(): Creating handler thread\n"));
    
    using_handler_thread = PORTS0_TRUE;
    
    thread = new_thread();
    thread->names = (ports0_thread_name_list_t *) NULL;
    ports0_thread_getspecific(_p0_context_key, &(thread->context));
    thread->user_func = handler_thread_func;
    thread->user_arg = NULL;
    
    while((rc = thr_create( NULL, 0, thread_starter, thread, 0, &handler_thread))==4);
    _p0_test_rc( rc, "PORTS0: create thread failed\n" );
    
} /* _p0_thread_create_handler_thread() */

/*
 * handler_thread_func()
 *
 * TODO: Should count 'unblocked_threads' as they call and return
 * from cond_wait.  Thus, when all threads except this one are blocked,
 * we know we can call _p0_blocking_poll() instead of ports0_poll when
 * using preemptive threads.
 */
static void *handler_thread_func(void *arg)
{
    ports0_debug_printf(2, ("handler_thread_func(): Entering\n"));

    while (!handler_thread_done)
    {
	    ports0_poll();
	    ports0_debug_printf(6, ("handler_thread_func(): calling yield\n"));
	    ports0_thread_yield();
	    ports0_debug_printf(6, ("handler_thread_func(): Has been scheduled\n"));
    }

    ports0_debug_printf(2,("handler_thread_func(): Exiting\n"));
    
    return (NULL);
} /* handler_thread_func() */

/*
 * _p0_thread_shutdown_handler_thread()
 */
void _p0_thread_shutdown_handler_thread()
{
    int rc;
    void *status;

    if (using_handler_thread)
    {
	handler_thread_done = PORTS0_TRUE;

	/*
	 * If this is not the idle thread, then wait for
	 * the idle thread to terminate.
	 */
	if ( thr_self() != handler_thread )
	{
	    rc = thr_join( handler_thread, NULL, &status );
	    _p0_test_rc( rc, "PORTS0: thread join failed\n" );
	}
    }
} /* _p0_thread_shutdown_handler_thread() */

/*
 * ports0_preemptive_threads
 *
 * Return PORTS0_TRUE (non-zero) if we are using preemptive threads.
 */
bool_t ports0_preemptive_threads()
{
    return (preemptive_threads);
} /* ports0_preemptive_threads() */


/*
 * ports0_key_create()
 */
int ports0_thread_key_create(ports0_thread_key_t *key,
			    ports0_thread_key_destructor_func_t func)
{
   int rc=0;

   while((rc = thr_keycreate( key, func )) == 4);
   _p0_test_rc( rc, "PORTS0: keycreate failed\n" );
    return(rc);
} /* ports0_key_create() */

/*
 * ports0_thread_setspecific()
 */
#undef ports0_thread_setspecific
void ports0_thread_setspecific(ports0_thread_key_t key,
			     void *value)
{
  ports0_macro_thread_setspecific(key, value);
} /* ports0_thread_setspecific() */

/*
 * ports0_thread_getspecific()
 */
#undef ports0_thread_getspecific
void ports0_thread_getspecific(ports0_thread_key_t key,
			     void **value)
{
  ports0_macro_thread_getspecific(key, value);
} /* ports0_thread_getspecific() */


/*
 * ports0_thread_self
 */
#undef ports0_thread_self
ports0_thread_t ports0_thread_self( void )
{
  return (ports0_macro_thread_self());
}

/*
 * ports0_thread_equal()
 */
#undef ports0_thread_equal
int ports0_thread_equal(ports0_thread_t t1,
               ports0_thread_t t2)
{
    return (ports0_macro_thread_equal(t1, t2));
} /* ports0_thread_equal() */

/*
 * ports0_thread_yield
 */
#undef ports0_thread_yield
void ports0_thread_yield( void )
{
  ports0_macro_thread_yield();
}

#undef ports0_mutex_init
void ports0_mutex_init( ports0_mutex_t *mut, ports0_mutexattr_t *attr )
{
  ports0_macro_mutex_init(mut, attr);
}

/*
 *  ports0_mutex_destroy()
 */
#undef ports0_mutex_destroy
void ports0_mutex_destroy( ports0_mutex_t *mut )
{
  ports0_macro_mutex_destroy(mut);
}

/*
 * ports0_cond_init()
 */
#undef ports0_cond_init
void ports0_cond_init( ports0_cond_t *cv, ports0_condattr_t *attr )
{
  ports0_macro_cond_init(cv, attr);
}

/*
 *  ports0_cond_destroy()
 */
#undef ports0_cond_destroy
void ports0_cond_destroy( ports0_cond_t *cv )
{
  ports0_macro_cond_destroy(cv);
}

/* 
 *  ports0_mutex_lock()
 */
#undef ports0_mutex_lock
void ports0_mutex_lock( ports0_mutex_t *mut )
{
  ports0_macro_mutex_lock(mut);
}

/*
 *  ports0_mutex_unlock()
 */
#undef ports0_mutex_unlock
void ports0_mutex_unlock( ports0_mutex_t *mut )
{
  ports0_macro_mutex_unlock(mut);
}

/*
 *  ports0_cond_wait()
 */
#undef ports0_cond_wait
void ports0_cond_wait( ports0_cond_t *cv, ports0_mutex_t *mut )
{
  ports0_macro_cond_wait(cv, mut);
}

/*
 *  ports0_cond_signal()
 */
#undef ports0_cond_signal
void ports0_cond_signal( ports0_cond_t *cv )
{
  ports0_macro_cond_signal(cv);
}

/*
 *  ports0_cond_broadcast()
 */
#undef ports0_cond_broadcast
void ports0_cond_broadcast( ports0_cond_t *cv )
{
  ports0_macro_cond_broadcast(cv);
}




/*
 * _p0_actual_thread_once()
 */
int _p0_actual_thread_once(ports0_thread_once_t *once_control,
			   void (*init_routine)())
{
    int rc;
    ports0_mutex_lock(&thread_once_mutex);
    if (*once_control)
    {
	/* Someone beat us to it.  */
	rc = 0;
    }
    else
    {
	/* We're the first one here */
	(*init_routine)();
	*once_control = 1;
	rc = 0;
    }
    ports0_mutex_unlock(&thread_once_mutex);
    return (rc);
} /* _p0_actual_thread_once() */

#undef ports0_thread_once
int ports0_thread_once(ports0_thread_once_t *once_control,
		        void (*init_routine)())
{
  return (_p0_thread_once(once_control, init_routine));
}

