/*
 * th_sol.h
 *
 * External header for Solaris threads
 *
 * rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_sol.h,v 1.1 1994/11/01 03:54:05 tuecke Exp $"
 */

/*
 * #defines
 * _P0_SYSTEM_COMPILE  -- we are compiling ports0 not user code
 * _P0_OPAQUE_HEADERS  -- hide the thread headers from the user
 * PORTS0_USE_HEADERS   -- allow Ports0 functions to be macros
 */


#include <thread.h>
#include <synch.h>

typedef mutex_t       _p0_mutex_t;
typedef cond_t        _p0_cond_t;
typedef thread_key_t  _p0_thread_key_t;
typedef int           _p0_thread_attr_t;
typedef int           _p0_mutexattr_t;
typedef int           _p0_condattr_t;

typedef size_t p0_size;
typedef void *(*p0_func)(void *);
typedef void * any_t;

#define _P0_THR_SUCCESS 0

typedef int                ports0_thread_t;
typedef _p0_thread_key_t   ports0_thread_key_t;
typedef int                ports0_thread_attr_t;
typedef int                ports0_mutexattr_t;
typedef int                ports0_condattr_t;

/*
 * ports0_mutex_t
 */
typedef struct _ports0_mutex_t {
  _P0_START_MAGIC_COOKIE
  _p0_mutex_t mutex;
  _P0_END_MAGIC_COOKIE
} ports0_mutex_t;

/*
 * ports0_cond_t
 */
typedef struct _ports0_cond_t {
  _P0_START_MAGIC_COOKIE
  _p0_cond_t cond;
  _P0_END_MAGIC_COOKIE
} ports0_cond_t;


#define PORTS0_THREAD_ONCE_INIT 0
typedef int ports0_thread_once_t;

typedef void (*ports0_thread_key_destructor_func_t)(void *value);

typedef struct _ports0_global_vars {
  _p0_thread_key_t        ports0_thread_t_pointer;
  int                     general_attribute;  
  int                     thread_flags;
} ports0_global_vars;

PORTS0_GLOBAL ports0_global_vars ports0_all_global_vars;

#ifdef __cplusplus
extern "C" {
#endif

/*
extern void _p0_more_thread_mem( void );
extern void _p0_init_thread_package( void );
extern void *_p0_thread_starter( void * );
*/

#ifdef __cplusplus
}
#endif

extern void _p0_report_bad_rc( int, char * );
extern int ports0_thread_create(ports0_thread_t *thread,
			       ports0_thread_attr_t *attr,
			       ports0_thread_func_t tar_func,
			       void *user_arg );
extern void ports0_thread_exit(void *status);

extern int ports0_thread_key_create(ports0_thread_key_t *key,
				   ports0_thread_key_destructor_func_t func);

#define _p0_thread_once( once_control, init_routine ) \
     (*once_control ? 0 : _p0_actual_thread_once(once_control, init_routine))

#define _p0_test_rc( a, b ) \
  do {                                                   \
    if( a != _P0_THR_SUCCESS && a != EINTR ) _p0_report_bad_rc( a, b ); \
  } while(0)

#define ports0_macro_thread_setspecific(key, value) \
do \
{ \
    int rc=0; \
              \
    while((rc = thr_setspecific(key, value))==4); \
    _p0_test_rc(rc, "PORTS0: ports0_thread_setspecific() failed\n"); \
} while(0)

#define ports0_macro_thread_getspecific(key, value) \
do \
{ \
    int rc=0;   \
                \
    while((rc = thr_getspecific(key, value))==4);  \
    _p0_test_rc(rc, "PORTS0: ports0_thread_getspecific() failed\n"); \
} while(0)

#define ports0_macro_thread_self() \
    thr_self()

#define ports0_macro_thread_equal( p0t1, p0t2 ) (p0t1 == p0t2)

#define ports0_macro_thread_once( once_control, init_routine ) \
     _p0_thread_once(once_control, init_routine)

#define ports0_macro_thread_yield()  \
    thr_yield()

#define ports0_macro_mutex_init( mut, attr ) \
  do {                                                      \
     int rc=0;                                              \
                                                            \
     _P0_INIT_START_MAGIC_COOKIE(mut);                      \
     while((rc = mutex_init( &((mut)->mutex), USYNC_THREAD, NULL ))==4); \
     _p0_test_rc( rc, "PORTS0: allocate lock failed\n" );    \
     _P0_INIT_END_MAGIC_COOKIE(mut);                        \
      } while(0)

#define ports0_macro_mutex_destroy( mut )                           \
 do {                                                        \
   int rc=0;                                                 \
                                                             \
   PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "destroy" );         \
   while((rc = mutex_destroy( &((mut)->mutex) ))==4);     \
   _p0_test_rc( rc, "PORTS0: free lock failed\n" );           \
 } while(0)

#define ports0_macro_cond_init( cv, attr )                           \
 do {                                                        \
   int rc=0;                                                 \
                                                             \
   _P0_INIT_START_MAGIC_COOKIE(cv);                          \
   while((rc = cond_init( &((cv)->cond), USYNC_THREAD, NULL))==4);     \
   _p0_test_rc( rc, "PORTS0: allocate condition variable failed\n"); \
   _P0_INIT_END_MAGIC_COOKIE(cv);                            \
 } while(0)

#define ports0_macro_cond_destroy( cv )                           \
 do {                                                            \
   int rc=0;                                                     \
                                                                 \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "destroy" );               \
   while((rc = cond_destroy( &((cv)->cond) ))==4) ;     \
   _p0_test_rc( rc, "PORTS0: free condition variable failed\n" ); \
 } while(0)

#define ports0_macro_mutex_lock( mut )                                     \
 do {                                                          \
  int rc=0;                                                    \
                                                               \
  PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock1" );              \
  while((rc = mutex_lock( &((mut)->mutex) ))==4) ;               \
  _p0_test_rc( rc, "PORTS0: mutex lock failed\n" );             \
  PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock2" );              \
 } while(0)

#define ports0_macro_mutex_unlock( mut )                                   \
 do {                                                          \
   int rc=0;                                                     \
                                                               \
   PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "unlock1" );           \
   while((rc = mutex_unlock( &((mut)->mutex) ))==4);             \
   _p0_test_rc( rc, "PORTS0: mutex unlock failed\n" );          \
 } while(0)

#define ports0_macro_cond_wait( cv, mut )                                   \
 do {                                                                      \
   int rc=0;                                                               \
								           \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "wait1" );                           \
   rc=cond_wait( &((cv)->cond), &((mut)->mutex) );                         \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "wait2" );                           \
   _p0_test_rc( rc, "PORTS0: condition variable wait failed\n" );           \
 } while(0)

#define ports0_macro_cond_signal( cv )                               \
 do {                                                               \
   int rc=0;                                                        \
                                                                    \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "signal1" );                  \
   while((rc = cond_signal( &((cv)->cond) ))==4) ;         \
   _p0_test_rc( rc, "PORTS0: condition variable signal failed\n" );  \
 } while(0)

#define ports0_macro_cond_broadcast( cv )                         \
 do {                                                               \
   int rc=0;                                                        \
                                                                    \
   PORTS0_INTERROGATE( cv, _P0_COND_T, "broadcast11" );                  \
   while((rc = cond_broadcast( &((cv)->cond) ))==4);       \
   _p0_test_rc( rc, "PORTS0: condition variable broadcast failed\n" );  \
 } while(0)


#ifdef PORTS0_USE_MACROS

#define ports0_thread_setspecific(key, value) \
    ports0_macro_thread_setspecific(key, value)
#define ports0_thread_getspecific(key, value) \
    ports0_macro_thread_getspecific(key, value)
#define ports0_thread_self() \
    ports0_macro_thread_self()
#define ports0_thread_equal(t1, t2) \
    ports0_macro_thread_equal(t1, t2)
#define ports0_thread_once(once_control, init_routine) \
    ports0_macro_thread_once(once_control, init_routine)
#define ports0_thread_yield() \
    ports0_macro_thread_yield()
#define ports0_mutex_init( mut, attr ) \
    ports0_macro_mutex_init( mut, attr )
#define ports0_mutex_destroy( mut ) \
    ports0_macro_mutex_destroy( mut )
#define ports0_mutex_lock( mut ) \
    ports0_macro_mutex_lock( mut )
#define ports0_mutex_unlock( mut ) \
    ports0_macro_mutex_unlock( mut )
#define ports0_cond_init( cv, attr ) \
    ports0_macro_cond_init( cv, attr )
#define ports0_cond_destroy( cv ) \
    ports0_macro_cond_destroy( cv )
#define ports0_cond_wait( cv, mut ) \
    ports0_macro_cond_wait( cv, mut )
#define ports0_cond_signal( cv ) \
    ports0_macro_cond_signal( cv )
#define ports0_cond_broadcast( cv ) \
    ports0_macro_cond_broadcast( cv )


#else  /* PORTS0_USE_MACROS */

extern void		ports0_thread_setspecific(ports0_thread_key_t key,
						 void *value);
extern void		ports0_thread_getspecific(ports0_thread_key_t key,
						 void **value);
extern ports0_thread_t	ports0_thread_self(void);
extern int		ports0_thread_equal(ports0_thread_t t1,
					   ports0_thread_t t2);
extern int		ports0_thread_once(ports0_thread_once_t *once_control,
					  void (*init_routine)());
extern void		ports0_thread_yield(void);
extern void		ports0_mutex_init(ports0_mutex_t *mutex,
					 ports0_mutexattr_t *attr);
extern void		ports0_mutex_destroy(ports0_mutex_t *mutex);
extern void		ports0_mutex_lock(ports0_mutex_t *mutex);
extern void		ports0_mutex_unlock(ports0_mutex_t *mutex);
extern void		ports0_cond_init(ports0_cond_t *cond,
					ports0_condattr_t *attr);
extern void		ports0_cond_destroy(ports0_cond_t *cond);
extern void		ports0_cond_wait(ports0_cond_t *cond,
					ports0_mutex_t *mutex);
extern void		ports0_cond_signal(ports0_cond_t *cond);
extern void		ports0_cond_broadcast(ports0_cond_t *cond);

#endif /* PORTS0_USE_MACROS */

